package de.fraunhofer.iem.dataprovider.kpi.service

import de.fraunhofer.iem.dataprovider.kpi.dto.KPIDto
import de.fraunhofer.iem.dataprovider.kpi.model.KPI
import de.fraunhofer.iem.dataprovider.kpi.model.KPIHierarchyEdge
import de.fraunhofer.iem.dataprovider.kpi.repository.KPIHierarchyEdgeRepository
import de.fraunhofer.iem.dataprovider.kpi.repository.KPIRepository
import de.fraunhofer.iem.dataprovider.logger.getLogger
import de.fraunhofer.iem.dataprovider.toolRun.model.Repository
import org.springframework.data.repository.findByIdOrNull
import org.springframework.stereotype.Service
import java.util.*

@Service
class KPIService(private val kpiRepository: KPIRepository, private val kpiHierarchyEdgeRepository: KPIHierarchyEdgeRepository) {

    private val logger = getLogger(javaClass)

    private fun calculateKPIsRecursively(kpi: KPIDto, visited: MutableSet<KPIDto>) {
        // Check if the KPI has already been visited
        if (visited.contains(kpi)) {
            return
        }

        // Check if the KPI has child KPIs
        if (kpi.hierarchyEdges.isEmpty()) {
            // Leaf node, calculate the KPI value
            kpi.calculateKPI()
            visited.add(kpi)
            return
        }

        // Recursively calculate child KPIs first
        for (childEdge in kpi.hierarchyEdges) {
            calculateKPIsRecursively(childEdge.to, visited)
        }

        // Calculate the KPI value after processing child KPIs
        kpi.calculateKPI()
        visited.add(kpi)
    }

    fun calculateKPIs(rootKPI: KPIDto) {
        val visited: MutableSet<KPIDto> = mutableSetOf()
        calculateKPIsRecursively(rootKPI, visited)
    }

    fun storeAndPurgeOld(repository: Repository, rootKPI: KPIDto) {
        purgeAllKPIs(repository)
        val kpiDtoToEntityMapping: MutableMap<KPIDto, KPI> = mutableMapOf()
        storeKPI(repository, rootKPI, kpiDtoToEntityMapping)
    }

    private fun storeKPI(repository: Repository, kpi: KPIDto, kpiDtoToEntityMapping: MutableMap<KPIDto, KPI>) {
        // already visited
        if (kpiDtoToEntityMapping.contains(kpi)) {
            return
        }

        if (kpi.hierarchyEdges.isEmpty()) {
            // leaf node
            val kpiEntity = kpi.toDbObject()
            kpiEntity.repository = repository

            kpiDtoToEntityMapping[kpi] = kpiEntity
            kpiRepository.save(kpiEntity)
            logger.info("Storing leaf node ${kpi.name}")
            return
        }

        for (childEdge in kpi.hierarchyEdges) {
            storeKPI(repository, childEdge.to, kpiDtoToEntityMapping)
        }

        val kpiEntity = kpi.toDbObject()
        kpiEntity.repository = repository
        kpiDtoToEntityMapping[kpi] = kpiEntity
        kpiRepository.save(kpiEntity)
        logger.info("Storing node ${kpi.name}")
        for (hierarchyEdge in kpi.hierarchyEdges) {
            val hierarchyEdgeEntity = KPIHierarchyEdge()
            hierarchyEdgeEntity.from = kpiEntity
            hierarchyEdgeEntity.to = kpiDtoToEntityMapping[hierarchyEdge.to]
            hierarchyEdgeEntity.weight = hierarchyEdge.weight
            kpiHierarchyEdgeRepository.save(hierarchyEdgeEntity)
            logger.info("Storing edge ${hierarchyEdge.weight}: From ${hierarchyEdge.from.name} to ${hierarchyEdge.to.name}")
        }
    }

    private fun purgeAllKPIs(repository: Repository) {
        val kpis: List<KPI> = kpiRepository.findByRepository_Id(repository.id!!)
        kpiRepository.deleteAll(kpis)
    }

    fun findKPIById(id: UUID): KPI? {
        return kpiRepository.findByIdOrNull(id)
    }

    fun findRootKPIByRepository(repository: Repository): KPI? {
        val kpi = kpiRepository.findFirstByRepository_IdAndIsRootTrue(repository.id!!)
        if (kpi.isEmpty) {
            return null
        }
        return kpi.get()
    }
}