package de.fraunhofer.iem.dataprovider.taskManager.tasks

import de.fraunhofer.iem.dataprovider.gitlab.GitConfiguration
import de.fraunhofer.iem.dataprovider.taskManager.model.Event
import de.fraunhofer.iem.dataprovider.taskManager.model.GetRepositoryDetailsDone
import de.fraunhofer.iem.dataprovider.toolRun.ToolRunService
import de.fraunhofer.iem.dataprovider.toolRun.model.Repository
import org.gitlab4j.api.GitLabApi
import org.gitlab4j.api.models.Commit
import org.gitlab4j.api.models.Project
import java.util.UUID

class GetRepositoryDetailsTask(
        private val repoId: UUID,
        private val gitlabConfiguration: GitConfiguration,
        private val toolRunService: ToolRunService,
        override val responseChannel: suspend (event: Event) -> Unit,
) : Task() {
    private val gitlabApi: GitLabApi = GitLabApi(gitlabConfiguration.host, gitlabConfiguration.accessToken)

    override suspend fun execute() {
        val repositoryEntity = toolRunService.findRepoByID(repoId);
        logger.info(repositoryEntity.toString())
        if (repositoryEntity != null) {
            val project = gitlabApi.projectApi.getProject(repositoryEntity.repoId)
            // Note: We only take commits from the default branch
            val commits = gitlabApi.commitsApi.getCommits(project.id, project.defaultBranch, ".")

            val numberOfCommits = commits.count()
            val numberOfSignedCommits = getNumberOfSignedCommits(repositoryEntity, commits)
            val isDefaultBranchProtected = isDefaultBranchProtected(repositoryEntity, project)
            val repositoryDetailsDto = RepositoryDetailsDto(repositoryEntity, numberOfCommits, numberOfSignedCommits, isDefaultBranchProtected)
            logger.info("Collected repository details from $repoId (${repositoryDetailsDto}) successfully")
            val repositoryDetailsEntity = toolRunService.createRepositoryDetails(repositoryDetailsDto);
            logger.info("Stored repository details for $repoId (${repositoryDetailsEntity}) successfully")
            responseChannel(GetRepositoryDetailsDone(taskID, repoId, repositoryDetailsEntity))
        } else {
            logger.error("Repository $repoId can not be found in the database")
        }
    }

    // TODO: This should probably live somewhere else and encapsulate the logic
    private fun getNumberOfSignedCommits(repositoryEntity: Repository, commits: List<Commit>): Int {
        var numberOfSignedCommits = 0
        for (commit in commits) {
            val signature = gitlabApi.commitsApi.getOptionalGpgSignature(repositoryEntity.repoId, commit.id)
            if (signature !== null) {
                numberOfSignedCommits++
            }
        }
        return numberOfSignedCommits
    }

    // TODO: This should probably live somewhere else and encapsulate the logic
    private fun isDefaultBranchProtected(repositoryEntity: Repository, project: Project): Boolean {
        return try {
            val defaultBranchName = project.defaultBranch;
            val branch = gitlabApi.repositoryApi.getBranch(repositoryEntity.repoId, defaultBranchName);
            branch.protected
        } catch (e: Exception) {
            // in theory, error probably happens if branch can't be found. In this case we default to false
            false;
        }
    }
}
